"use strict";
/*
    Migrate.js - OneTable Migrations
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Migrate = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const semver_1 = __importDefault(require("semver"));
const dynamodb_onetable_1 = require("dynamodb-onetable");
// DEV
// import { Model, Table } from '../../onetable/dist/mjs/index.js'
const MigrationKey = '_Migration';
class Migrate {
    /*
        Construct the Migrate instance.
        The `config` parameter is either instance of Table() or params for Table.
        The `params.migrations` set to the in-memory migrations or `params.dir` to a migrations directory.
     */
    constructor(config = {}, params = {}) {
        this.params = params;
        this.migrations = params.migrations;
        this.dir = path_1.default.resolve(params.dir || '.');
        this.db = (typeof config.setClient != 'function') ? new dynamodb_onetable_1.Table(config) : config;
    }
    /*
        Initialize and define standard models: Schema, Migration
    */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.db.setSchema();
        });
    }
    /* public */
    apply(direction, version) {
        return __awaiter(this, void 0, void 0, function* () {
            let db = this.db;
            let migration, model;
            console.log(`Apply migration ${version} direction ${direction}`);
            if (direction == 0) {
                migration = yield this.loadMigration('latest');
                model = db.getModel(MigrationKey);
                //  Remove all existing migrations
                yield model.remove({}, { many: true });
                //  Create prior migration items (not last)
                let versions = yield this.getVersions();
                version = versions.pop();
                for (let v of versions) {
                    let migration = yield this.loadMigration(v);
                    model = db.getModel(MigrationKey);
                    yield model.create({
                        date: new Date(),
                        description: migration.description,
                        path: migration.path,
                        version: migration.version,
                    });
                }
            }
            if (!migration) {
                migration = yield this.loadMigration(version);
                model = db.getModel(MigrationKey);
            }
            if (direction < 0) {
                yield migration.down(db, this);
                yield model.remove({ version: migration.version });
                let current = yield this.loadMigration(yield this.getCurrentVersion());
                yield db.saveSchema(current.schema);
            }
            else {
                //  Up, repeat or reset
                yield migration.up(db, this);
                db.saveSchema(migration.schema);
                let params = {
                    version,
                    date: new Date(),
                    path: migration.path,
                    description: migration.description,
                };
                if (direction <= 1) {
                    yield model.create(params);
                }
                else {
                    yield model.update(params);
                }
            }
        });
    }
    /* public */
    findPastMigrations() {
        return __awaiter(this, void 0, void 0, function* () {
            let model = yield this.db.getModel(MigrationKey);
            let pastMigrations = yield model.find({});
            this.sortMigrations(pastMigrations);
            return pastMigrations;
        });
    }
    /* public */
    getCurrentVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            let pastMigrations = yield this.findPastMigrations();
            if (pastMigrations.length == 0) {
                return '0.0.0';
            }
            return pastMigrations[pastMigrations.length - 1].version;
        });
    }
    /*
        Return outstanding versions in semver sorted order up to the specified limit
     */
    /* public */
    getOutstandingVersions(limit = Number.MAX_SAFE_INTEGER) {
        return __awaiter(this, void 0, void 0, function* () {
            let current = yield this.getCurrentVersion();
            let pastMigrations = yield this.findPastMigrations();
            let versions = yield this.getVersions();
            versions = versions.filter(version => {
                return semver_1.default.compare(version, current) > 0;
            }).sort(semver_1.default.compare);
            versions = versions.filter(v => pastMigrations.find(m => m.version == v) == null);
            return versions.slice(0, limit);
        });
    }
    /*
        Return outstanding migrations in semver sorted order up to the specified limit
     */
    /* public */
    getOutstandingMigrations(limit = Number.MAX_SAFE_INTEGER) {
        return __awaiter(this, void 0, void 0, function* () {
            let versions = yield this.getOutstandingVersions(limit);
            let migrations = [];
            for (let v of versions) {
                migrations.push(yield this.loadMigration(v));
            }
            return migrations;
        });
    }
    /* private */
    getVersions() {
        return __awaiter(this, void 0, void 0, function* () {
            let versions;
            if (this.migrations) {
                versions = this.migrations.map(m => m.version);
            }
            else {
                versions = fs_1.default.readdirSync(this.dir).map(file => file.replace(/\.[^/.]+$/, ''));
            }
            return versions.filter(version => { return semver_1.default.valid(version); }).sort(semver_1.default.compare);
        });
    }
    /* private */
    /*
        Load a migration and set its schema. Return the migration (modified)
    */
    loadMigration(version) {
        return __awaiter(this, void 0, void 0, function* () {
            let path, migration;
            if (this.migrations) {
                migration = this.migrations.find(m => m.version == version);
            }
            else {
                path = `${this.dir}/${version}.js`;
                migration = (yield Promise.resolve().then(() => __importStar(require(path)))).default;
            }
            if (!migration) {
                throw new Error(`Cannot find migration for version ${version}`);
            }
            if (!migration.schema) {
                throw new Error(`Migration ${version} is missing a schema`);
            }
            if (!migration.version) {
                throw new Error(`Migration ${version} is missing a version property`);
            }
            yield this.db.setSchema(migration.schema);
            return {
                description: migration.description,
                down: migration.down,
                path: path || 'memory',
                schema: migration.schema,
                up: migration.up,
                version,
            };
        });
    }
    /* private */
    sortMigrations(array) {
        array.sort((a, b) => {
            let cmp = semver_1.default.compare(a.version, b.version);
            if (cmp < 0) {
                return cmp;
            }
            else if (cmp > 0) {
                return cmp;
            }
            else if (a.order < b.order) {
                return -1;
            }
            else if (a.order > b.order) {
                return 1;
            }
            else {
                return 0;
            }
        });
    }
}
exports.Migrate = Migrate;
